/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/* 
 * ezfc.c
 * Copyright (C) 2011-2012 Akira TAGOH
 * 
 * Authors:
 *   Akira TAGOH  <akira@tagoh.org>
 * 
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_TRUETYPE_TABLES_H
#include "ezfc.h"

typedef enum {
	EZFC_FONT_UNKNOWN   = 0,
	EZFC_FONT_SANS      = 1 << 0,
	EZFC_FONT_SERIF     = 1 << 1,
	EZFC_FONT_MONOSPACE = 1 << 2,
	EZFC_FONT_CURSIVE   = 1 << 3,
	EZFC_FONT_FANTASY   = 1 << 4,
	EZFC_FONT_END
} ezfc_font_t;
typedef GList * (ezfc_proc_t) (GList *, const FcPattern *);

static gboolean __ezfc_is_initialized = FALSE;
static FT_Library __ezfc_ftlib = NULL;

/*< private >*/
static FT_Face
_ezfc_get_face(const FcPattern *pattern)
{
	FcChar8 *file;
	int idx;
	FT_Face face = NULL;

	FcPatternGetString(pattern, FC_FILE, 0, &file);
	FcPatternGetInteger(pattern, FC_INDEX, 0, &idx);

	if (FT_New_Face(__ezfc_ftlib, (char *)file, idx, &face)) {
		g_warning("Unable to open the font file '%s' index %d",
			  file, idx);
	}

	return face;
}

static ezfc_font_t
_ezfc_get_font_type(const gchar *font)
{
	static const gchar *aliases[] = {
		"sans",
		"sans-serif",
		"serif",
		"monospace",
		"cursive",
		"fantasy",
		NULL
	};
	gint i;

	for (i = 0; aliases[i] != NULL; i++) {
		if (g_ascii_strcasecmp(font, aliases[i]) == 0) {
			if (i == 0)
				i++;
			return 1 << (i - 1);
		}
	}

	return EZFC_FONT_UNKNOWN;
}

static ezfc_font_t
_ezfc_get_font_type_from_pattern(const FcPattern *pattern)
{
	int spacing = 0;
	ezfc_font_t retval = EZFC_FONT_UNKNOWN;

	FcPatternGetInteger(pattern, FC_SPACING, 0, &spacing);
	if (spacing == FC_MONO) {
		retval = EZFC_FONT_MONOSPACE;
	} else {
		TT_OS2 *os2;
		FT_Face face = NULL;

		face = _ezfc_get_face(pattern);
		os2 = (TT_OS2 *)FT_Get_Sfnt_Table(face, ft_sfnt_os2);
		if (!os2)
			goto bail;
		/* See http://www.monotypeimaging.com/ProductsServices/pan1.aspx
		 * for the reference of Panose
		 */
		if (os2->panose[0] == 3) {
			/* Latin Hand Written */
			retval |= EZFC_FONT_CURSIVE;
		} else if (os2->panose[0] >= 4 &&
			   os2->panose[0] <= 5) {
			/* 4...Latin Decorative
			 * 5...Latin Symbol
			 */
			retval |= EZFC_FONT_FANTASY;
		} else if (os2->panose[0] == 2 &&
			   ((os2->panose[1] >= 11 &&
			     os2->panose[1] <= 13) ||
			    os2->panose[1] == 15)) {
			/* 2..Latin Text */
			/*   11...Normal Sans
			 *   12...Obtuse Sans
			 *   13...Perpendicular Sans
			 *   15...Pounded
			 */
			retval |= EZFC_FONT_SANS;
		} else if (os2->panose[0] == 1) {
			/* "Not Fit" in the family kind should be ignored. */
		} else if (os2->panose[0] == 0 &&
			   os2->panose[1] == 1) {
			/* Unable to determine the alias type */
		} else {
			retval |= EZFC_FONT_SERIF;
		}
		if ((os2->panose[0] == 2 &&
		     os2->panose[3] == 9) ||
		    (os2->panose[0] == 3 &&
		     os2->panose[3] == 3) ||
		    (os2->panose[0] == 4 &&
		     os2->panose[3] == 9) ||
		    (os2->panose[0] == 5 &&
		     os2->panose[3] == 3)) {
			retval |= EZFC_FONT_MONOSPACE;
		}
	  bail:
		FT_Done_Face(face);
	}

	return retval;
}

static gboolean
_ezfc_is_font_a(const FcPattern *pattern,
		const gchar     *alias_name)
{
	ezfc_font_t type = _ezfc_get_font_type(alias_name);
	ezfc_font_t ptype = _ezfc_get_font_type_from_pattern(pattern);

	return (type & ptype) == type;
}

static GList *
_ezfc_get_fonts_proc(const gchar *language,
		     const gchar *alias_name,
		     ezfc_proc_t  func)
{
	FcPattern *pat;
	FcFontSet *fs;
	FcObjectSet *os;
	GList *retval = NULL;
	FcLangSet *ls = NULL;

	g_return_val_if_fail (func != NULL, NULL);

	if (alias_name) {
		g_return_val_if_fail (ezfc_is_alias_font(alias_name), NULL);
	}
	pat = FcPatternCreate();
	if (language) {
		ls = FcLangSetCreate();
		FcLangSetAdd(ls, (const FcChar8 *)language);
		FcPatternAddLangSet(pat, FC_LANG, ls);
	}
	os = FcObjectSetBuild(FC_FAMILY, FC_FILE, FC_INDEX, FC_SPACING, NULL);
	fs = FcFontList(NULL, pat, os);
	FcObjectSetDestroy(os);
	FcPatternDestroy(pat);
	if (ls)
		FcLangSetDestroy(ls);
	if (fs) {
		gint i;

		for (i = 0; i < fs->nfont; i++) {
			if (!alias_name || _ezfc_is_font_a(fs->fonts[i], alias_name)) {
				retval = func(retval, fs->fonts[i]);
			}
		}
		FcFontSetDestroy(fs);
	}

	return retval;
}

static GList *
_ezfc_get_fonts_cb_family(GList           *list,
			  const FcPattern *pattern)
{
	FcChar8 *family;
	GList *l;
	gboolean added = FALSE;

	FcPatternGetString(pattern, FC_FAMILY, 0, &family);
	for (l = list; l != NULL; l = g_list_next(l)) {
		gint r = g_strcmp0(l->data, (const gchar *)family);
		GList *ll;

		if (r == 0) {
			/* do not add it */
			added = TRUE;
			break;
		} else if (r > 0) {
			ll = g_list_prepend(l, g_strdup((const gchar *)family));

			if (list == l)
				list = ll;
			added = TRUE;
			break;
		}
	}
	if (!added)
		list = g_list_append(list, g_strdup((const gchar *)family));

	return list;
}

static GList *
_ezfc_get_fonts_cb_pattern(GList           *list,
			   const FcPattern *pattern)
{
	return g_list_append(list, FcPatternDuplicate(pattern));
}

/*< public >*/

/**
 * ezfc_error_get_quark:
 *
 * Obtains a #GQuark being used for #GError.
 *
 * Returns: a #GQuark.
 */
GQuark
ezfc_error_get_quark(void)
{
	static GQuark quark = 0;

	if (quark == 0)
		quark = g_quark_from_static_string("ezfc-error-quark");

	return quark;
}

/**
 * ezfc_init:
 *
 * Initialize the library.
 */
void
ezfc_init(void)
{
	if (!__ezfc_is_initialized) {
		__ezfc_is_initialized = TRUE;
		FT_Init_FreeType(&__ezfc_ftlib);
		FcInit();
	}
}

/**
 * ezfc_finalize:
 *
 * Finalize the library.
 */
void
ezfc_finalize(void)
{
	if (__ezfc_is_initialized) {
		__ezfc_is_initialized = FALSE;
		FcFini();
		FT_Done_FreeType(__ezfc_ftlib);
		__ezfc_ftlib = NULL;
	}
}

/**
 * ezfc_is_alias_font:
 * @alias_name: the alias font name
 *
 * Checks if @alias_name is one of sans-serif, serif, monospace, cursive or fantasy.
 *
 * Returns: %TRUE if @alias_name is an alias font name, otherwise %FALSE.
 */
gboolean
ezfc_is_alias_font(const gchar *alias_name)
{
	ezfc_font_t t = _ezfc_get_font_type(alias_name);

	return t != EZFC_FONT_UNKNOWN;
}

/**
 * ezfc_get_fonts_list:
 * @language: (allow-none): the language name fontconfig can deal with.
 * @alias_name: (allow-none): the alias name to obtain the fonts list for.
 *
 * Obtains the fonts list being assigned to @alias_name for @language.
 *
 * Returns: (element-type utf8) (transfer full): a #GList contains the font family name.
 *          if no valid families, %NULL then.
 */
GList *
ezfc_get_fonts_list(const gchar *language,
		    const gchar *alias_name)
{
	return _ezfc_get_fonts_proc(language, alias_name,
				    _ezfc_get_fonts_cb_family);
}

/**
 * ezfc_get_fonts_pattern_list:
 * @language: (allow-none): the language name fontconfig can deal with.
 * @alias_name: (allow-none): the alias name to obtain the fonts pettern list for.
 *
 * Obtains #FcPattern list being assigned to @alias_name for @language.
 *
 * Returns: (element-type FcPattern) (transfer full): a #GList contains #FcPattern, otherwise %NULL.
 */
GList *
ezfc_get_fonts_pattern_list(const gchar *language,
			    const gchar *alias_name)
{
	return _ezfc_get_fonts_proc(language, alias_name,
				    _ezfc_get_fonts_cb_pattern);
}
